import math
import os

import pygame

from .OneClickManager import OneClickManager


class Label:
    labels = []
    def __init__(
        self,
        surface: pygame.Surface,
        text,
        size: int,
        xy: tuple,
        anchor="center",
        **kwargs,
    ):
        """
        Creates a label.

        Instructions:
            - To create a label, create an instance of this class before
            the mainloop of the game.
            - To make the label appear, call the method 'self.draw()' in
            every loop of the game.

        Example (simplified):
            label = Label(screen, "Hello world!", 32, (100,100), "topleft")
            while True:
                label.draw()

        Arguments:
            surface
                the surface the object will be drawn onto
                Type: pygame.Surface
            text
                the text displayed on the label / button
                Type: Any
            size
                refers to the size of text in pixels
                Type: int
            xy
                refers to the position of the anchor
                Type:
                    tuple[int, int]
                    tuple[tuple[int, int], tuple[int, int]] -> performs sum of vectors (v1[0] + v2[0], v1[1] + v2[1])
                    list[...]
                    list[list[...]]
            anchor
                the anchor of the object:   topleft,    midtop,    topright,
                                            midleft,    center,    midright,
                                            bottomleft, midbottom, bottomright
                Type: str
            textcolor
                color of the text in rgb
                Type: tuple, list
            backgroundcolor
                color of the background in rgb
                Type: None, tuple, list
            antialias
                whether antialias (blurring to make text seem more realistic) should be used
                Type: bool
            font
                The font used for the text. Tries to match a font of PygameXtras first, then checks sysfonts.
                To get a list of all available PygameXtras fonts, run 'PygameXtras.list_fonts()'. To get a list
                of all sysfonts, run 'pygame.font.get_fonts()'.
                Type: str
            x_axis_addition
                adds the given amount of pixels to the left and right (-> x axis) of the text
                Type: int
            y_axis_addition
                adds the given amount of pixels to the top and bottom (-> y axis) of the text
                Type: int
            borderwidth
                width of the border
                Type: int
            bordercolor
                color of the border
                Type: tuple, list
            force_width
                forces the width of the background
                Type: int
            force_height
                forces the height of the background
                Type: int
            force_dim
                forces width and height of the background
                Type: tuple, list
            borderradius
                used to round the corners of the background_rect; use int for all corners or other types for individual corners
                Type: int, tuple, list
            text_offset
                moves the text by the given values; is inverted if binding_rect == 1
                Type: tuple, list
            image
                takes an image as background for the widget; configures the widget to use the dimensions of the image, unless specified otherwise
                Type: pygame.Surface
            info
                can store any kind of data; usage of dict recommended
                Type: Any
            text_binding
                binds the text within the background_rect
                Type: str (same as <anchor>)
            highlight
                (button only)
                makes the button shine when the mouse hovers over it; when using "True", the button shines a little brighter than
                its usual backgroundcolor; otherwise, the button shines in the given rgb color; only works when using <backgroundcolor>
                Type: True, tuple, list
            active_area
                (button only)
                makes the self.update method only work if the mouse click / hovering is also within the specified rect; useful if
                buttons are moveable but should not be clickable if they are outside a certain area
                Type: tuple, list, pygame.Rect
            bold
                makes the text appear bold (faked)
                Type: bool
            italic
                makes the text appear italic (faked)
                Type: bool
            underline
                underlines the text
                Type: bool
            one_click_manager
                (button only)
                assures that overlaying buttons can not be clicked at the same time
                Type: PygameXtras.OneClickManager
            padding (NOT IMPLEMENTED YET)
                decreases the size of the widget without affecting the position
                Type: int
            font_file
                use font from a file (path expected)
                Type: str

        All custom arguments can also be used in their short form (eg. "aa" instead of "antialias").
        To see what all the short forms look like, inspect the self.ABBREVIATIONS attribute.
        """

        self.ABBREVIATIONS = {
            "textcolor": "tc",
            "backgroundcolor": "bgc",
            "antialias": "aa",
            "font": "f",
            "x_axis_addition": "xad",
            "y_axis_addition": "yad",
            "borderwidth": "bw",
            "bordercolor": "bc",
            "force_width": "fw",
            "force_height": "fh",
            "force_dim": "fd",
            "borderradius": "br",
            "text_offset": "to",
            "image": "img",
            "info": "info",
            "text_binding": "tb",
            "highlight": "hl",
            "active_area": "aA",
            "bold": "bo",
            "italic": "it",
            "underline": "ul",
            "one_click_manager": "ocm",
            "template": "t",
            "padding": "p",
            "font_file": "ff",
        }
        assert len(set(self.ABBREVIATIONS.values())) == len(
            list(self.ABBREVIATIONS.values())
        ), "it seems like two arguments share the same abbreviation"

        # inserting template (if exists)
        template = kwargs.get("template", None)
        if template is None:
            template = kwargs.get(self.ABBREVIATIONS["template"], None)
        if template is not None:
            self.__test_type("template", template, dict)
            for k in template.keys():
                if not k in kwargs.keys():
                    kwargs[k] = template[k]

        self.__is_touching__ = False  # if cursor is touching the rect, only for buttons
        self.__has_hl_image__ = False

        self.surface = surface
        self.text = str(text)

        self.size = size
        self.__test_type("size", size, (int, float))
        self.size = int(self.size)

        self.__test_type("xy", xy, (tuple, list))
        self.__test_len("xy", xy, "=", 2)
        try:
            if isinstance(xy[0], (int, float)) and isinstance(xy[1], (int, float)):
                self.xy = int(xy[0]), int(xy[1])
            elif isinstance(xy[0], (tuple, list)) and isinstance(xy[1], (tuple, list)):
                t1 = xy[0]
                t2 = xy[1]
                self.__test_len("xy", t1, "=", 2)
                self.__test_len("xy", t1, "=", 2)
                self.__test_type("xy", t1[0], (int, float))
                self.__test_type("xy", t1[1], (int, float))
                self.__test_type("xy", t2[0], (int, float))
                self.__test_type("xy", t2[1], (int, float))
                self.xy = (int(t1[0] + t2[0]), int(t1[1] + t2[1]))
            else:
                self.__collect_error("xy", self.xy)
        except:
            self.__collect_error("xy", self.xy)

        self.anchor = anchor
        if "anchor" in kwargs.keys():
            self.anchor = kwargs["anchor"]
        self.__test_match(
            "anchor",
            self.anchor,
            "topleft,midtop,topright,midleft,center,midright,bottomleft,midbottom,bottomright".split(
                ","
            ),
        )

        kw = kwargs

        # textcolor
        self.textcolor = kw.get("textcolor", None)
        if self.textcolor == None:
            self.textcolor = kw.get(self.ABBREVIATIONS["textcolor"], None)
        if self.textcolor == None:
            self.textcolor = (0, 0, 0)
        # assertion
        if self.textcolor != None:
            self.__test_type("textcolor", self.textcolor, (tuple, list))
            self.__test_len("textcolor", self.textcolor, "=", 3)

        # backgroundcolor
        self.backgroundcolor = kw.get("backgroundcolor", None)
        if self.backgroundcolor == None:
            self.backgroundcolor = kw.get(self.ABBREVIATIONS["backgroundcolor"], None)
        # assertion
        if self.backgroundcolor != None:
            self.__test_type("backgroundcolor", self.backgroundcolor, (tuple, list))
            self.__test_len("backgroundcolor", self.backgroundcolor, "=", 3)

        # backgroundcolor backup
        self.backgroundcolor_init = self.backgroundcolor
        # assertion completed by previous assertion

        # antialias
        self.antialias = kw.get("antialias", None)
        if self.antialias == None:
            self.antialias = kw.get(self.ABBREVIATIONS["antialias"], None)
        if self.antialias == None:
            self.antialias = True
        # assertion
        self.antialias = bool(self.antialias)

        # font
        self.font = kw.get("font", None)
        if self.font == None:
            self.font = kw.get(self.ABBREVIATIONS["font"], None)
        if self.font == None:
            self.font = ""
        # assertion
        if self.font != None:
            self.__test_type("font", self.font, str)

        # font_file
        self.font_file = kw.get("font_file", None)
        if self.font_file == None:
            self.font_file = kw.get(self.ABBREVIATIONS["font_file"], None)
        # assertion
        if self.font_file != None:
            self.__test_type("font_file", self.font_file, str)

        # x_axis_addition
        self.x_axis_addition = kw.get("x_axis_addition", None)
        if self.x_axis_addition == None:
            self.x_axis_addition = kw.get(self.ABBREVIATIONS["x_axis_addition"], None)
        if self.x_axis_addition == None:
            self.x_axis_addition = 0
        # assertion
        if self.x_axis_addition != None:
            self.__test_type("x_axis_addition", self.x_axis_addition, int)
            self.__test_value("x_axis_addition", self.x_axis_addition, ">=", 0)

        # y_axis_addition
        self.y_axis_addition = kw.get("y_axis_addition", None)
        if self.y_axis_addition == None:
            self.y_axis_addition = kw.get(self.ABBREVIATIONS["y_axis_addition"], None)
        if self.y_axis_addition == None:
            self.y_axis_addition = 0
        # assertion
        if self.y_axis_addition != None:
            self.__test_type("y_axis_addition", self.y_axis_addition, int)
            self.__test_value("y_axis_addition", self.y_axis_addition, ">=", 0)

        # borderwidth
        self.borderwidth = kw.get("borderwidth", None)
        if self.borderwidth == None:
            self.borderwidth = kw.get(self.ABBREVIATIONS["borderwidth"], None)
        if self.borderwidth == None:
            self.borderwidth = 0
        # assertion
        if self.borderwidth != None:
            self.__test_type("borderwidth", self.borderwidth, int)
            self.__test_value("borderwidth", self.borderwidth, ">=", 0)

        # bordercolor
        self.bordercolor = kw.get("bordercolor", None)
        if self.bordercolor == None:
            self.bordercolor = kw.get(self.ABBREVIATIONS["bordercolor"], None)
        if self.bordercolor == None:
            self.bordercolor = (0, 0, 0)
        # assertion
        if self.bordercolor != None:
            self.__test_type("bordercolor", self.bordercolor, (tuple, list))
            self.__test_len("bordercolor", self.bordercolor, "=", 3)

        # force_width
        self.force_width = kw.get("force_width", None)
        if self.force_width == None:
            self.force_width = kw.get(self.ABBREVIATIONS["force_width"], None)
        # assertion
        if self.force_width != None:
            self.__test_type("force_width", self.force_width, int)
            self.__test_value("force_width", self.force_width, ">", 0)

        # force_height
        self.force_height = kw.get("force_height", None)
        if self.force_height == None:
            self.force_height = kw.get(self.ABBREVIATIONS["force_height"], None)
        # assertion
        if self.force_height != None:
            self.__test_type("force_height", self.force_height, int)
            self.__test_value("force_height", self.force_height, ">", 0)

        # force_dim
        force_dim = kw.get("force_dim", None)
        if force_dim == None:
            force_dim = kw.get(self.ABBREVIATIONS["force_dim"], None)
        # assertion
        if force_dim != None:
            self.__test_type("force_dim", force_dim, (tuple, list))
            self.__test_len("force_dim", force_dim, "=", 2)
            if force_dim[0] != None:
                self.force_width = force_dim[0]
            if force_dim[1] != None:
                self.force_height = force_dim[1]

        # borderradius
        self.borderradius = kw.get("borderradius", None)
        if self.borderradius == None:
            self.borderradius = kw.get(self.ABBREVIATIONS["borderradius"], None)
        if self.borderradius == None:
            self.borderradius = (1, 1, 1, 1)
        # assertion
        elif type(self.borderradius) == int:
            self.borderradius = (
                self.borderradius,
                self.borderradius,
                self.borderradius,
                self.borderradius,
            )
        elif type(self.borderradius) in [tuple, list]:
            self.__test_len("borderradius", self.borderradius, "=", 4)
        else:
            self.__collect_error("borderradius", self.borderradius)

        # text_offset
        self.text_offset = kw.get("text_offset", None)
        if self.text_offset == None:
            self.text_offset = kw.get(self.ABBREVIATIONS["text_offset"], None)
        if self.text_offset == None:
            self.text_offset = (0, 0)
        # assertion
        if self.text_offset != None:
            self.__test_type("text_offset", self.text_offset, (tuple, list))
            self.__test_len("text_offset", self.text_offset, "=", 2)

        # image
        self.image = kw.get("image", None)
        if self.image == None:
            self.image = kw.get(self.ABBREVIATIONS["image"], None)
        # assertion
        if self.image != None:
            self.__test_type("image", self.image, pygame.Surface)
            if self.force_width == None:
                self.force_width = self.image.get_width()
            if self.force_height == None:
                self.force_height = self.image.get_height()
            self.image = pygame.transform.scale(
                self.image, (self.force_width, self.force_height)
            )

        # info
        self.info = kw.get("info", None)
        if self.info == None:
            self.info = kw.get(self.ABBREVIATIONS["info"], None)
        # no assertion needed

        # text_binding
        self.text_binding = kw.get("text_binding", None)
        if self.text_binding == None:
            self.text_binding = kw.get(self.ABBREVIATIONS["text_binding"], None)
        if self.text_binding == None:
            self.text_binding = "center"
        # assertion
        if self.text_binding != None:
            self.__test_match(
                "text_binding",
                self.text_binding,
                "topleft,midtop,topright,midleft,center,midright,bottomleft,midbottom,bottomright".split(
                    ","
                ),
            )

        # highlight
        self.highlight = kw.get("highlight", None)
        if self.highlight == None:
            self.highlight = kw.get(self.ABBREVIATIONS["highlight"], None)
        # assertion
        if self.highlight != None:
            if type(self.highlight) in [tuple, list] and len(self.highlight) == 3:
                pass  # ok
            elif self.highlight == False:
                self.highlight = None  # ok
            elif self.highlight == True:
                if self.image == None:
                    if not self.backgroundcolor != None:
                        self.__collect_error_directly(
                            f"'backgroundcolor' (currently: {self.backgroundcolor}) must be defined when using 'highlight' (currently: {self.highlight})"
                        )
            else:
                self.__collect_error("highlight", self.highlight)

            if self.image == None:
                if self.highlight == True:
                    self.highlight = (
                        min(self.backgroundcolor[0] + 50, 255),
                        min(self.backgroundcolor[1] + 50, 255),
                        min(self.backgroundcolor[2] + 50, 255),
                    )
            else:
                self.__has_hl_image__ = True
                # if the backgroundcolor is not defined, the regular image will be taken for the hl_image
                # if there is a backgroundcolor, the backgroundcolor will be included in the hl_image
                if self.backgroundcolor == None:
                    self.hl_image = self.image.copy()
                else:
                    self.hl_image = pygame.Surface(self.image.get_size())
                    self.hl_image.fill(self.backgroundcolor)
                    self.hl_image.blit(self.image, (0, 0))
                # applying the brightening effect / coloring
                if self.highlight == True:
                    self.hl_image.fill((50, 50, 50), special_flags=pygame.BLEND_RGB_ADD)
                else:
                    self.hl_image.fill(
                        self.highlight, special_flags=pygame.BLEND_RGB_MULT
                    )
                # scaling to fit the size
                self.hl_image = pygame.transform.scale(
                    self.hl_image, (self.force_width, self.force_height)
                )

        # active_area
        self.active_area = kw.get("active_area", None)
        if self.active_area == None:
            self.active_area = kw.get(self.ABBREVIATIONS["active_area"], None)
        # assertion
        if self.active_area != None:
            self.__test_type(
                "active_area", self.active_area, (tuple, list, pygame.Rect)
            )
            self.__test_len("active_area", self.active_area, "=", 4)
            self.active_area = pygame.Rect(self.active_area)

        # bold
        self.bold = kw.get("bold", None)
        if self.bold == None:
            self.bold = kw.get(self.ABBREVIATIONS["bold"], None)
        if self.bold == None:
            self.bold = False
        # assertion
        self.bold = bool(self.bold)

        # italic
        self.italic = kw.get("italic", None)
        if self.italic == None:
            self.italic = kw.get(self.ABBREVIATIONS["italic"], None)
        if self.italic == None:
            self.italic = False
        # assertion
        self.italic = bool(self.italic)

        # underline
        self.underline = kw.get("underline", None)
        if self.underline == None:
            self.underline = kw.get(self.ABBREVIATIONS["underline"], None)
        if self.underline == None:
            self.underline = False
        # assertion
        self.underline = bool(self.underline)

        # one_click_manager
        self.one_click_manager = kw.get("one_click_manager", None)
        if self.one_click_manager == None:
            self.one_click_manager = kw.get(
                self.ABBREVIATIONS["one_click_manager"], None
            )
        # assertion
        if self.one_click_manager != None:
            self.__test_type(
                "one_click_manager", self.one_click_manager, OneClickManager
            )

        # padding
        self.padding = kw.get("padding", None)
        if self.padding == None:
            self.padding = kw.get(self.ABBREVIATIONS["padding"], None)
        if self.padding == None:
            self.padding = 0
        # assertion
        if self.padding != None:
            self.__test_type("padding", self.padding, int)
            self.__test_value("padding", self.padding, ">=", 0)

        self.__load_font_path()
        self.__create__()

    def __create__(self):

        print(f"{self.font_path = }")
        font = pygame.font.Font(self.font_path, self.size)
        font.set_bold(self.bold)
        font.set_italic(self.italic)
        font.set_underline(self.underline)
        self.text_surface = font.render(str(self.text), self.antialias, self.textcolor)
        self.text_rect = self.text_surface.get_rect()

        # x y w h
        # x_axis: x, w
        # y_axis: y, h

        if self.x_axis_addition == 0:
            x = self.text_rect.x
            w = self.text_rect.width
        elif self.x_axis_addition > 0:
            x = self.text_rect.x - self.x_axis_addition
            w = self.text_rect.width + self.x_axis_addition * 2

        if self.y_axis_addition == 0:
            y = self.text_rect.y
            h = self.text_rect.height
        elif self.y_axis_addition > 0:
            y = self.text_rect.y - self.y_axis_addition
            h = self.text_rect.height + self.y_axis_addition * 2

        if self.force_width == None:
            pass
        else:
            w = self.force_width

        if self.force_height == None:
            pass
        else:
            h = self.force_height

        # creating the background rect
        self.background_rect = pygame.Rect(
            x + self.padding,
            y + self.padding,
            w - 2 * self.padding,
            h - 2 * self.padding,
        )

        # creating positioning rect
        self.positioning_rect = pygame.Rect(x, y, w, h)

        # putting everything in correct position
        self.update_pos(self.xy, self.anchor)

    def __load_font_path(self):
        if self.font_file is None:
            fonts = [
                f.removesuffix(".ttf")
                for f in os.listdir(
                    os.path.join(os.path.dirname(__file__), "..", "fonts")
                )
            ]
            if self.font.lower() in fonts:
                self.font_path = os.path.join(
                    os.path.dirname(__file__), "..", "fonts", self.font.lower() + ".ttf"
                )
            else:
                self.font_path = pygame.font.match_font(self.font)
        else:
            self.font_path = self.font_file

    def __test_type(self, variable_name, variable_value, applicable_types):
        if not isinstance(variable_value, applicable_types):
            raise ValueError(
                f"invalid argument for '{variable_name}': {variable_value}"
            )

    def __test_len(self, variable_name, variable_value, _type: str, value):
        assert _type in ">,<,=,>=,<=".split(",")
        if (
            _type == "<"
            and not len(variable_value) < value
            or _type == "<="
            and not len(variable_value) <= value
            or _type == ">"
            and not len(variable_value) > value
            or _type == ">="
            and not len(variable_value) >= value
            or _type == "="
            and not len(variable_value) == value
        ):
            raise ValueError(
                f"invalid argument for '{variable_name}': {variable_value}"
            )

    def __test_value(self, variable_name, variable_value, _type: str, value):
        assert _type in ">,<,=,>=,<=".split(",")
        if (
            _type == "<"
            and not variable_value < value
            or _type == "<="
            and not variable_value <= value
            or _type == ">"
            and not variable_value > value
            or _type == ">="
            and not variable_value >= value
            or _type == "="
            and not variable_value == value
        ):
            raise ValueError(
                f"invalid argument for '{variable_name}': {variable_value}"
            )

    def __test_match(self, variable_name, variable_value, applicable_values: tuple):
        assert isinstance(applicable_values, (tuple, list))
        if not variable_value in applicable_values:
            raise ValueError(
                f"invalid argument for '{variable_name}': {variable_value}"
            )

    def draw(self):
        """
        Draws the widget to the screen.
        """

        if self.backgroundcolor != None:
            if self.__is_touching__ and self.image == None:
                pygame.draw.rect(
                    self.surface,
                    self.highlight,
                    self.background_rect,
                    0,
                    border_top_left_radius=self.borderradius[0],
                    border_top_right_radius=self.borderradius[1],
                    border_bottom_right_radius=self.borderradius[2],
                    border_bottom_left_radius=self.borderradius[3],
                )
            else:
                pygame.draw.rect(
                    self.surface,
                    self.backgroundcolor,
                    self.background_rect,
                    0,
                    border_top_left_radius=self.borderradius[0],
                    border_top_right_radius=self.borderradius[1],
                    border_bottom_right_radius=self.borderradius[2],
                    border_bottom_left_radius=self.borderradius[3],
                )
        if self.image != None:
            if self.__has_hl_image__ and self.__is_touching__:
                self.surface.blit(self.hl_image, self.background_rect)
            else:
                self.surface.blit(self.image, self.background_rect)

        if self.borderwidth > 0:
            pygame.draw.rect(
                self.surface,
                self.bordercolor,
                self.background_rect,
                self.borderwidth,
                border_top_left_radius=self.borderradius[0],
                border_top_right_radius=self.borderradius[1],
                border_bottom_right_radius=self.borderradius[2],
                border_bottom_left_radius=self.borderradius[3],
            )
        self.surface.blit(self.text_surface, self.text_rect)

    def draw_to(self, surface: pygame.Surface):
        """
        Draws the widget to a different surface than initially specified.
        """
        assert (
            type(surface) == pygame.Surface
        ), f"invalid argument for 'surface': {surface}"

        if self.backgroundcolor != None:
            if self.__is_touching__ and self.image == None:
                pygame.draw.rect(
                    surface,
                    self.highlight,
                    self.background_rect,
                    0,
                    border_top_left_radius=self.borderradius[0],
                    border_top_right_radius=self.borderradius[1],
                    border_bottom_right_radius=self.borderradius[2],
                    border_bottom_left_radius=self.borderradius[3],
                )
            else:
                pygame.draw.rect(
                    surface,
                    self.backgroundcolor,
                    self.background_rect,
                    0,
                    border_top_left_radius=self.borderradius[0],
                    border_top_right_radius=self.borderradius[1],
                    border_bottom_right_radius=self.borderradius[2],
                    border_bottom_left_radius=self.borderradius[3],
                )
        if self.image != None:
            if self.__has_hl_image__ and self.__is_touching__:
                surface.blit(self.hl_image, self.background_rect)
            else:
                surface.blit(self.image, self.background_rect)

        if self.borderwidth > 0:
            pygame.draw.rect(
                surface,
                self.bordercolor,
                self.background_rect,
                self.borderwidth,
                border_top_left_radius=self.borderradius[0],
                border_top_right_radius=self.borderradius[1],
                border_bottom_right_radius=self.borderradius[2],
                border_bottom_left_radius=self.borderradius[3],
            )
        surface.blit(self.text_surface, self.text_rect)

    def update_text(self, text):
        """
        Updates the text of a widget. Call this
        method before drawing to the screen.
        """
        if str(text) != str(self.text):
            self.text = str(text)
            self.__create__()

    def update_colors(self, textcolor=None, backgroundcolor=None, bordercolor=None):
        """
        Updates the colors of a widget. Call this
        method before drawing to the screen.

        This method can be called every loop without worrying about performance problems.
        """
        if textcolor != None and textcolor != self.textcolor:
            if not type(textcolor) in [tuple, list]:
                raise ValueError(
                    f"invalid argument for 'textcolor': {textcolor}"
                )
            if not len(textcolor) == 3:
                raise ValueError(f"invalid argument for 'textcolor': {textcolor}")
            self.textcolor = textcolor
            self.__create__()
        if backgroundcolor != None and backgroundcolor != self.backgroundcolor:
            if not type(backgroundcolor) in [tuple, list]:
                raise ValueError(f"invalid argument for 'backgroundcolor': {backgroundcolor}")
            if not len(backgroundcolor) == 3:
                raise ValueError(f"invalid argument for 'backgroundcolor': {backgroundcolor}")
            self.backgroundcolor = backgroundcolor
            self.backgroundcolor_init = backgroundcolor
            val = 50
            self.highlight = (
                min(self.backgroundcolor[0] + val, 255),
                min(self.backgroundcolor[1] + val, 255),
                min(self.backgroundcolor[2] + val, 255),
            )
            self.__create__()
        if bordercolor != None and bordercolor != self.bordercolor:
            if not type(bordercolor) in [tuple, list]:
                raise ValueError(f"invalid argument for 'bordercolor': {bordercolor}")
            if not len(bordercolor) == 3:
                raise ValueError(f"invalid argument for 'bordercolor': {bordercolor}")
            self.bordercolor = bordercolor

    def update_borderwidth(self, borderwidth: int):
        """
        Updates the borderwidth of the widget. Call
        this method before drawing to the screen.
        """
        if not type(borderwidth) == int:
            raise ValueError(f"invalid argument for 'borderwidth': {borderwidth}")
        if not borderwidth >= 0:
            raise ValueError(f"invalid argument for 'borderwidth': {borderwidth}")
        self.borderwidth = borderwidth

    def update_pos(self, xy, anchor=None):
        """
        Changes the widgets position. Call this
        method before drawing to the screen.
        """
        if not isinstance(xy, (tuple, list)):
            raise ValueError(f"invalid argument for 'xy': {xy}")
        if not len(xy) == 2:
            raise ValueError(f"invalid argument for 'xy': {xy}")
        self.xy = xy
        if not anchor in [
            "topleft",
            "topright",
            "bottomleft",
            "bottomright",
            "center",
            "midtop",
            "midright",
            "midbottom",
            "midleft",
            None,
        ]:
            raise ValueError(f"invalid argument for 'anchor': {anchor}")
        if anchor is not None:
            self.anchor = anchor

        self.positioning_rect.__setattr__(self.anchor, self.xy)
        self.background_rect.__setattr__("center", self.positioning_rect.center)
        background_rect_coords = getattr(self.background_rect, self.text_binding)
        self.text_rect.__setattr__(
            self.text_binding,
            (
                background_rect_coords[0] + self.text_offset[0],
                background_rect_coords[1] + self.text_offset[1],
            ),
        )

        # data # should actually be accessed through 'self.rect.' but I
        # will leave it as it is to not break any programs
        self.topleft = self.positioning_rect.topleft
        self.topright = self.positioning_rect.topright
        self.bottomleft = self.positioning_rect.bottomleft
        self.bottomright = self.positioning_rect.bottomright
        self.center = self.positioning_rect.center
        self.midtop = self.positioning_rect.midtop
        self.midright = self.positioning_rect.midright
        self.midbottom = self.positioning_rect.midbottom
        self.midleft = self.positioning_rect.midleft
        self.left = self.positioning_rect.left
        self.right = self.positioning_rect.right
        self.top = self.positioning_rect.top
        self.bottom = self.positioning_rect.bottom

        self.rect = self.positioning_rect

        # for buttons:
        self.x_range = (
            self.background_rect.x,
            self.background_rect.x + self.background_rect.width,
        )
        self.y_range = (
            self.background_rect.y,
            self.background_rect.y + self.background_rect.height,
        )

    def set_style(self, bold: bool = None, italic: bool = None, underline: bool = None):
        old_bold = self.bold
        if bold != None:
            self.bold = bool(bold)
        old_italic = self.italic
        if italic != None:
            self.italic = bool(italic)
        old_underline = self.underline
        if underline != None:
            self.underline = bool(underline)
        if (
            old_bold != self.bold
            or old_italic != self.italic
            or old_underline != self.underline
        ):
            self.__create__()

    def get_rect(self) -> pygame.Rect:
        return self.rect



import pygame
import sys

class PerformanceGraph:
    def __init__(self, desired_fps: int, fps_list: list[float]):
        win_w, win_h = 1200, 600
        border = 50


        pygame.init()
        screen = pygame.display.set_mode((win_w, win_h))
        fpsclock = pygame.time.Clock()
        fps = 60

        pygame.draw.line(screen, (0,0,255), (border, 100), (win_w - border, 100), 1)
        pygame.draw.line(screen, (255,255,255), (border, border), (border, win_h-border), 1)
        pygame.draw.line(screen, (255,255,255), (border, win_h - border), (win_w - border, win_h-border), 1)

        x = [2+border+i*((win_w-2*border)/len(fps_list)) for i in range(len(fps_list))]
        y = [win_h-border-((win_h-border*3)/desired_fps)*item for item in fps_list]
        # for x_, y_ in zip(x, y):
        #     pygame.draw.circle(screen, (255,0,0), (x_, y_), 1)
        for x_, y_ in zip(range(len(x)-1), range(len(y)-1)):
            pygame.draw.line(screen, (255,0,0), (x[x_], y[y_]), (x[x_+1], y[y_+1]))
        

        pygame.display.flip()

        run = True
        while run:
            event_list = pygame.event.get()
            for event in event_list:
                if event.type == pygame.QUIT:
                    pygame.quit()
                    sys.exit()
                if event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_ESCAPE:
                        run = False

            fpsclock.tick(fps)



import pygame

class Keyboard:
    def __init__(self):
        """
        Makes it easier to take user input (via the keyboard)
        """
        self.keys = {
            pygame.K_a: "a",
            pygame.K_b: "b",
            pygame.K_c: "c",
            pygame.K_d: "d",
            pygame.K_e: "e",
            pygame.K_f: "f",
            pygame.K_g: "g",
            pygame.K_h: "h",
            pygame.K_i: "i",
            pygame.K_j: "j",
            pygame.K_k: "k",
            pygame.K_l: "l",
            pygame.K_m: "m",
            pygame.K_n: "n",
            pygame.K_o: "o",
            pygame.K_p: "p",
            pygame.K_q: "q",
            pygame.K_r: "r",
            pygame.K_s: "s",
            pygame.K_t: "t",
            pygame.K_u: "u",
            pygame.K_v: "v",
            pygame.K_w: "w",
            pygame.K_x: "x",
            pygame.K_y: "y",
            pygame.K_z: "z",
            pygame.K_0: "0",
            pygame.K_1: "1",
            pygame.K_2: "2",
            pygame.K_3: "3",
            pygame.K_4: "4",
            pygame.K_5: "5",
            pygame.K_6: "6",
            pygame.K_7: "7",
            pygame.K_8: "8",
            pygame.K_9: "9",
            pygame.K_TAB: "\t",
            pygame.K_SPACE: " ",
            pygame.K_SEMICOLON: ";",
            pygame.K_SLASH: "/",
            pygame.K_RETURN: "\n",
            pygame.K_PERIOD: ".",
            pygame.K_PLUS: "+",
            pygame.K_PERCENT: "%",
            pygame.K_MINUS: "-",
            pygame.K_HASH: "#",
            pygame.K_UNDERSCORE: "_",
            pygame.K_LEFTBRACKET: "(",
            pygame.K_RIGHTBRACKET: ")",
            pygame.K_LESS: "<",
            pygame.K_GREATER: ">",
            pygame.K_EQUALS: "=",
            pygame.K_EURO: "€",
            pygame.K_EXCLAIM: "!",
            pygame.K_QUESTION: "?",
            pygame.K_DOLLAR: "$",
            pygame.K_COLON: ":",
            pygame.K_COMMA: ",",
            pygame.K_BACKSLASH: "\\",
            pygame.K_ASTERISK: "*",
        }
        self.shift_keys = {
            pygame.K_a: "A",
            pygame.K_b: "B",
            pygame.K_c: "C",
            pygame.K_d: "D",
            pygame.K_e: "E",
            pygame.K_f: "F",
            pygame.K_g: "G",
            pygame.K_h: "H",
            pygame.K_i: "I",
            pygame.K_j: "J",
            pygame.K_k: "K",
            pygame.K_l: "L",
            pygame.K_m: "M",
            pygame.K_n: "N",
            pygame.K_o: "O",
            pygame.K_p: "P",
            pygame.K_q: "Q",
            pygame.K_r: "R",
            pygame.K_s: "S",
            pygame.K_t: "T",
            pygame.K_u: "U",
            pygame.K_v: "V",
            pygame.K_w: "W",
            pygame.K_x: "X",
            pygame.K_y: "Y",
            pygame.K_z: "Z",
            pygame.K_0: "=",
            pygame.K_1: "!",
            pygame.K_2: "\"",
            pygame.K_3: "§",
            pygame.K_4: "$",
            pygame.K_5: "%",
            pygame.K_6: "&",
            pygame.K_7: "/",
            pygame.K_8: "(",
            pygame.K_9: ")",
            pygame.K_PERIOD: ":",
            pygame.K_PLUS: "*",
            pygame.K_MINUS: "_",
            pygame.K_HASH: "'",
            pygame.K_LESS: ">",
            pygame.K_COMMA: ";"
        }
        self.__forbidden_chars__ = []

    def set_custom_value(self, custom_values: dict):
        """
        Sets custom return values specified by <custom_values>.
        """
        assert type(custom_values) == dict
        for k, v in custom_values.items():
            self.keys[k] = v
            self.shift_keys[k] = v

    def set_forbidden_characters(self, characters: list):
        """
        Bans all given characters.
        """
        assert type(characters) == list, f"invalid argument for 'characters': {characters}"
        for char in characters:
            assert type(char) == str, f"invalid character '{char}'"
            assert len(char) == 1, f"invalid character '{char}'"
            if char not in self.__forbidden_chars__:
                self.__forbidden_chars__.append(char)

    def get(self, event_list):
        """
        Returns a string of all keys pressed.
        """
        string = ""
        keys = pygame.key.get_pressed()
        # if keys[pygame.locals.K_LSHIFT] or keys[pygame.locals.K_RSHIFT]:
        if keys[pygame.K_LSHIFT] or keys[pygame.K_RSHIFT]:
            for e in event_list:
                if e.type == pygame.KEYDOWN and e.key:
                    char = self.shift_keys.get(e.key, "")
                    if char not in self.__forbidden_chars__:
                        string += char
        else:
            for e in event_list:
                if e.type == pygame.KEYDOWN and e.key:
                    char = self.keys.get(e.key, "")
                    if char not in self.__forbidden_chars__:
                        string += char
        return string



import pygame
from ..Label import Label


def up(size: int):
    points = []

    angle = 110

    v = pygame.Vector2(0, -1)
    v.scale_to_length(size * 0.4)
    points.append((v.x, v.y))

    v.scale_to_length(size * 0.55)
    v.rotate_ip(angle)
    points.append((v.x, v.y))

    v.rotate_ip((180 - angle) * 2)
    points.append((v.x, v.y))

    return points


def right(size: int):
    points = []

    angle = 110

    v = pygame.Vector2(1, 0)
    v.scale_to_length(size * 0.4)
    points.append((v.x, v.y))

    v.scale_to_length(size * 0.55)
    v.rotate_ip(angle)
    points.append((v.x, v.y))

    v.rotate_ip((180 - angle) * 2)
    points.append((v.x, v.y))

    return points


def down(size: int):
    points = []

    angle = 110

    v = pygame.Vector2(0, 1)
    v.scale_to_length(size * 0.4)
    points.append((v.x, v.y))

    v.scale_to_length(size * 0.55)
    v.rotate_ip(angle)
    points.append((v.x, v.y))

    v.rotate_ip((180 - angle) * 2)
    points.append((v.x, v.y))

    return points


def left(size: int):
    points = []

    angle = 110

    v = pygame.Vector2(-1, 0)
    v.scale_to_length(size * 0.4)
    points.append((v.x, v.y))

    v.scale_to_length(size * 0.55)
    v.rotate_ip(angle)
    points.append((v.x, v.y))

    v.rotate_ip((180 - angle) * 2)
    points.append((v.x, v.y))

    return points


def triangle(size: int):
    points = []

    v = pygame.Vector2(0, -1)
    v.scale_to_length(size * 0.6)
    points.append((v.x, v.y))

    v.rotate_ip(120)
    points.append((v.x, v.y))

    v.rotate_ip(120)
    points.append((v.x, v.y))

    return points


def square(size: int):
    points = []

    v = pygame.Vector2(1, 1)
    v.scale_to_length(size * 0.6)
    for _ in range(4):
        v.rotate_ip(90)
        points.append((v.x, v.y))

    return points


def slash(size: int):
    points = []

    v = pygame.Vector2(0, -1)
    v.scale_to_length(size * 1.2)
    v.rotate_ip(25)
    points.append((v.x, v.y))
    v.rotate_ip(180)
    points.append((v.x, v.y))

    return points


def __small_box(size: int):
    points = []

    v = pygame.Vector2(0, -1)
    v.scale_to_length(size * 1)

    v.rotate_ip(60)
    points.append((v.x, v.y))

    v.rotate_ip(60)
    points.append((v.x, v.y))

    v.rotate_ip(120)
    points.append((v.x, v.y))

    v.rotate_ip(60)
    points.append((v.x, v.y))

    return points


def __big_box(size: int):
    points = []

    v = pygame.Vector2(0, -1)
    v.scale_to_length(size * 1)

    angle_start = 35
    angle_between = 95

    v.rotate_ip(angle_start)
    points.append((v.x, v.y))

    v.rotate_ip(angle_between)
    points.append((v.x, v.y))

    v = pygame.Vector2(0, -1)
    v.scale_to_length(size * 1)

    v.rotate_ip(-angle_start)
    points.append((v.x, v.y))

    v.rotate_ip(-angle_between)
    points.insert(2, (v.x, v.y))

    return points


size_factor = 0.6


def l1(size: int, color: tuple):
    return __small_box(size), Label(
        None, "L1", int(size * size_factor), (0, 0), tc=color, f="verdana"
    )


def l2(size: int, color: tuple):
    return __big_box(size), Label(
        None, "L2", int(size * size_factor), (0, 0), tc=color, f="verdana"
    )


def l3(size: int, color: tuple):
    return Label(None, "L3", int(size * size_factor), (0, 0), tc=color, f="verdana")


def r1(size: int, color: tuple):
    return __small_box(size), Label(
        None, "R1", int(size * size_factor), (0, 0), tc=color, f="verdana"
    )


def r2(size: int, color: tuple):
    return __big_box(size), Label(
        None, "R2", int(size * size_factor), (0, 0), tc=color, f="verdana"
    )


def r3(size: int, color: tuple):
    return Label(None, "R3", int(size * size_factor), (0, 0), tc=color, f="verdana")


def __octagon(size: int):
    points = []

    v = pygame.Vector2(0, 1)
    v.scale_to_length(size)

    corners = 8

    v.rotate_ip(360 / corners / 2)

    for _ in range(corners):
        points.append((v.x, v.y))
        v.rotate_ip(360 / corners)

    return points


def ps(size: int, color: tuple):
    return __octagon(size), Label(
        None, "PS", int(size * size_factor), (0, 0), tc=color, f="verdana"
    )


def options(size: int, color: tuple):
    return __octagon(size), Label(
        None, "OPT", int(size * size_factor), (0, 0), tc=color, f="verdana"
    )


def share(size: int, color: tuple):
    return __octagon(size), Label(
        None, "SHA", int(size * size_factor), (0, 0), tc=color, f="verdana"
    )



import pygame
from .Label import Label

class Button(Label):
    def __init__(self, surface, text, size, xy: tuple, anchor="center", **kwargs):
        """
        Creates a clickable button.

        Instructions:
            - To create a button, create an instance of this class before
            the mainloop of the game.
            - To make the button appear, call the method 'self.draw()' in
            every loop of the game.
            - To check if the button has been clicked, call the method
            'self.update()' with a list of mouse events in an if-statement.

        Example (simplified):
            button = Button(screen, "Hello world!", 32, (100,100), "topleft")
            while True:
                events = pygame.event.get()
                if button.update(events):
                    ...
                button.draw()

        For information about arguments see parent class (Label).
        """

        super().__init__(surface, text, size, xy, anchor, **kwargs)

    def update(self, event_list, button: int = 1, offset: tuple = (0, 0)) -> bool:
        """
        Checks if button has been pressed and returns True if so.
        <button> can specify a certain button (1-3).
        <offset> will be subtracted from the current mouse position.
        """
        assert type(button) == int, f"invalid argument for 'button': {button}"
        assert 1 <= button <= 5, f"invalid argument for 'button': {button}"
        assert type(offset) in [tuple, list], f"invalid argument for 'offset': {offset}"
        assert len(offset) == 2, f"invalid argument for 'offset: {offset}"

        # stops if one_click_manager tells that a click has taken place elsewhere
        if self.one_click_manager != None and self.one_click_manager.get_clicked() == True:
            self.__is_touching__ = False
            return False

        # managing the actual clicks
        for event in event_list:
            if event.type == pygame.MOUSEBUTTONUP:
                pos = list(event.pos)
                if self.active_area != None and not self.active_area.collidepoint(pos):
                    continue
                pos[0] -= offset[0]
                pos[1] -= offset[1]
                if button == None:
                    if self.x_range[0] < pos[0] < self.x_range[1] and self.y_range[0] < pos[1] < self.y_range[1]:
                        if self.one_click_manager != None:
                            self.one_click_manager.set_clicked()
                        return True

                elif 0 < button < 4:
                    if button == event.button:
                        if self.x_range[0] < pos[0] < self.x_range[1] and self.y_range[0] < pos[1] < self.y_range[1]:
                            if self.one_click_manager != None:
                                self.one_click_manager.set_clicked()
                            return True

                else:
                    raise ValueError(f"invalid argument for 'button': {button}")

        if self.one_click_manager != None and self.one_click_manager.get_hovering() == True:
            self.__is_touching__ = False
            return False

        # managing the hovering (highlight)
        pos = list(pygame.mouse.get_pos())
        if self.active_area != None and not self.active_area.collidepoint(pos):
            self.__is_touching__ = False
            return False
        pos[0] -= offset[0]
        pos[1] -= offset[1]
        if self.x_range[0] < pos[0] < self.x_range[1] and self.y_range[0] < pos[1] < self.y_range[1]:
            if self.highlight != None:
                self.__is_touching__ = True
            if self.one_click_manager != None:
                self.one_click_manager.set_hovering()
        else:
            self.__is_touching__ = False

        return False



import pygame
from .Button import Button
from .Keyboard import Keyboard

class Entry(Button):
    def __init__(self, surface, text, size, xy: tuple, anchor="center", **kwargs):
        """
        Click to activate, click somewhere else to deactivate.
        If active, typed letters will appear on the widget.
        An indication, that the entry widget is active, has to
        be implemented externally (use self.get_state()).

        IMPORTANT:
            - The update method takes a complete event list as
              an input, while a regular button just needs a mouse event list.
            - Entries (unlike buttons) should NOT be updated in an if-statement!

        Additional keywords:
            max_chars / mc
                maximum number of characters
                Type: int
            text_when_empty / twe
                text that appears when there is currently no other content
                Type: str
            auto_style / ast
                if text_when_empty is set, this text will appear italic,
                while the real user input will appear in the regular font style
                IMPORTANT: this mechanic breaks if the style is set manually
                after the first initialization of the widget
                Type: bool
            strict_input / si
                only allow numeric ("int", "float", 0 - +inf by using "int+" / "float+") or alphabetic ("str") input
                Type: str
            show_cursor
                show a moveable (left and right) cursor, that makes it possible
                to edit text in the middle without deleting everything to its right
                Type: bool
        """

        # max_chars
        self.max_chars = kwargs.get("max_chars", None)
        if self.max_chars == None:
            self.max_chars = kwargs.get("max", None)
        # assertion
        if self.max_chars != None:
            assert type(self.max_chars) == int, f"invalid argument for 'max_chars': {self.max_chars}"
            assert self.max_chars >= 0, f"invalid argument for 'max_chars': {self.max_chars}"

        # min_chars
        self.min_chars = kwargs.get("min_chars", None)
        if self.min_chars == None:
            self.min_chars = kwargs.get("min", None)
        # assertion
        if self.min_chars != None:
            assert type(self.min_chars) == int, f"invalid argument for 'min_chars': {self.min_chars}"
            assert self.min_chars >= 0, f"invalid argument for 'min_chars': {self.min_chars}"

        # text_when_empty
        self.text_when_empty = kwargs.get("text_when_empty", None)
        if self.text_when_empty == None:
            self.text_when_empty = kwargs.get("twe", None)
        if self.text_when_empty != None:
            self.text_when_empty = str(self.text_when_empty)
        # assertion not needed

        # auto_style
        self.auto_style = kwargs.get("auto_style", None)
        if self.auto_style == None:
            self.auto_style = kwargs.get("ast", None)
        if self.auto_style == None:
            self.auto_style = False
        # assertion
        self.auto_style = bool(self.auto_style)

        # strict_input
        self.strict_input = kwargs.get("strict_input", None)
        if self.strict_input == None:
            self.strict_input = kwargs.get("si", None)
        # assertion
        assert self.strict_input is None or self.strict_input in ["int", "float", "int+", "float+", "str"], f"invalid argument for 'strict_input': {self.strict_input}"

        # show_cursor
        self.show_cursor = kwargs.get("show_cursor", None)
        if self.show_cursor == None:
            self.show_cursor = kwargs.get("sc", True)
        # assertion
        self.show_cursor = bool(self.show_cursor)
        self.__cursor_pos = 0 # 0 means "at the end", while 2 means "2 characters from the end", ...
        self.__old_cursor_pos = 0
        
        
        

        self.__state__ = False
        self.__old_state__ = False
        self.__force__ = False # used to set state even before the update method
        self.__permanent_state__ = None
        self.__value__ = str(text)
        self.__old_value__ = str(text)

        self.__keyboard__ = Keyboard()
        self.__keyboard__.set_forbidden_characters(["\t", "\n"])




        super().__init__(surface, text, size, xy, anchor, **kwargs)

        self.bold_init = self.bold
        self.italic_init = self.italic

        self.__manage_twe()

    def update(self, event_list, button: int = 1, offset: tuple = (0, 0)) -> bool:
        """
        Checks if entry has been clicked on and activates widget if so.
        Should be used with a regular event_list.
        <button> can specify a certain button (1-3).
        Also updates the text, if input is detected.
        """
        assert type(button) == int, f"invalid argument for 'button': {button}"
        assert 1 <= button <= 5, f"invalid argument for 'button': {button}"
        assert type(offset) in [tuple, list], f"invalid argument for 'offset': {offset}"
        assert len(offset) == 2, f"invalid argument for 'offset: {offset}"

        # stops if one_click_manager tells that a click has taken place elsewhere
        if self.one_click_manager != None and self.one_click_manager.get_clicked() == True:
            self.__is_touching__ = False
            return False
        
        # managing the actual clicks
        for event in event_list:
            if event.type == pygame.MOUSEBUTTONUP:
                pos = list(event.pos)
                if self.active_area != None and not self.active_area.collidepoint(pos):
                    continue
                pos[0] -= offset[0]
                pos[1] -= offset[1]
                if button == None:
                    if self.x_range[0] < pos[0] < self.x_range[1] and self.y_range[0] < pos[1] < self.y_range[1]:
                        if self.one_click_manager != None:
                            self.one_click_manager.set_clicked()
                        self.__state__ =  True
                    else:
                        self.__state__ = False

                elif 0 < button < 4:
                    if button == event.button:
                        if self.x_range[0] < pos[0] < self.x_range[1] and self.y_range[0] < pos[1] < self.y_range[1]:
                            if self.one_click_manager != None:
                                self.one_click_manager.set_clicked()
                            self.__state__ =  True
                        else:
                            self.__state__ = False

                else:
                    raise ValueError(f"invalid argument for 'button': {button}")
        if self.__permanent_state__ != None:
            self.__state__ = self.__permanent_state__

        if self.__force__ != None:
            self.__state__ = self.__force__
            self.__force__ = None

        # managing the hovering (highlight)
        pos = list(pygame.mouse.get_pos())
        if self.active_area != None and not self.active_area.collidepoint(pos):
            self.__is_touching__ = False
            self.__state__ = False
        pos[0] -= offset[0]
        pos[1] -= offset[1]
        if self.x_range[0] < pos[0] < self.x_range[1] and self.y_range[0] < pos[1] < self.y_range[1]:
            if self.highlight != None:
                self.__is_touching__ = True
            if self.one_click_manager != None:
                self.one_click_manager.set_hovering()
        else:
            self.__is_touching__ = False

        if self.__state__:

            # deleting chars
            for event in event_list:
                if event.type == pygame.KEYDOWN and event.key == pygame.K_BACKSPACE:
                    if self.min_chars == None or len(self.__value__) - 1 >= self.min_chars:
                        if self.show_cursor:
                            if self.__cursor_pos != len(self.__value__):
                                val_list = list(self.__value__)
                                del val_list[len(self.__value__)-1-self.__cursor_pos]
                                self.__value__ = "".join(val_list)
                        else:
                            self.__value__ = self.__value__[:-1]

                # managing cursor
                if self.show_cursor and event.type == pygame.KEYUP:
                    if event.key == pygame.K_LEFT:
                        self.__cursor_pos += 1
                    elif event.key == pygame.K_RIGHT:
                        self.__cursor_pos -= 1
                    elif event.key == 1073741898: # key "pos1"
                        self.__cursor_pos = len(self.__value__)
                    elif event.key == 1073741901: # key "ende"
                        self.__cursor_pos = 0

            # adding chars
            val = self.__keyboard__.get(event_list)
            new_val_list: list = list(self.__value__)
            new_val_list.insert(len(self.__value__) - self.__cursor_pos, val)
            new_val: str = "".join(new_val_list)

            # checks for self.strict_input and sets val to "" if condition does not apply
            if self.strict_input is not None and val != "":

                if new_val == "-":
                    if not (self.strict_input == "int+" or self.strict_input == "float+"):
                        self.__value__ = new_val
                elif new_val.startswith("-"):
                    if (self.strict_input == "int"     and new_val[1:].isnumeric()) or \
                       (self.strict_input == "float"   and "".join(new_val[1:].split(".", 1)).isnumeric()) or \
                       (self.strict_input == "str"     and new_val[1:].isalpha()):
                        self.__value__ = new_val
                else:
                    if (self.strict_input == "int"     and new_val.isnumeric()) or \
                       (self.strict_input == "float"   and "".join(new_val.split(".", 1)).isnumeric()) or \
                       (self.strict_input == "str"     and new_val.isalpha()) or \
                       (self.strict_input == "int+"    and new_val.isnumeric()) or \
                       (self.strict_input == "float+"  and "".join(new_val.split(".", 1)).isnumeric()):
                        self.__value__ = new_val
                        
            else:
                self.__value__ = new_val



            # dealing with self.max_chars
            if self.max_chars is not None:
                self.__value__ = self.__value__[:self.max_chars]

        if self.__value__ != self.__old_value__ or self.__cursor_pos != self.__old_cursor_pos or self.__old_state__ != self.__state__:
            if self.show_cursor:
                if self.__old_state__ != self.__state__:
                    self.__cursor_pos = 0
                else:
                    self.__clamp_cursor()
                self.__old_cursor_pos = self.__cursor_pos
            self.__old_state__ = self.__state__
            self.__refresh_text()

            # maybe the following line has to be unindented once, I am not sure
            self.__manage_twe()

    def get_state(self):
        """
        Returns boolean whether the entry is active (True) or not (False).
        """
        return self.__state__

    def get(self):
        """
        Returns the text.
        """
        return self.__value__

    def set(self, value):
        """
        Sets the value of the Entry.
        """
        self.__value__ = str(value)
        self.__refresh_text()

    def set_forbidden_characters(self, characters: list):
        """
        Bans all given characters.
        """
        self.__keyboard__.set_forbidden_characters(characters)

    def set_forbidden_characters_for_filename(self):
        forbidden_chars_in_filenames = "<>:\"/\\|?* \n\t\r,.;"
        self.set_forbidden_characters(list(forbidden_chars_in_filenames))

    def clear(self):
        """
        Clears the text.
        """
        self.__value__ = ""
        self.__refresh_text()

    def set_state(self, state: bool):
        """
        Sets activity state.
        """
        self.__force__ = bool(state)

    def set_permanent_state(self, state: bool):
        """
        Sets activity state permanently. Remove with self.remove_permanent_state().
        """
        self.__permanent_state__ = bool(state)

    def remove_permanent_state(self):
        """
        Removes permanent state.
        """
        self.__permanent_state__ = None
    
    def __refresh_text(self):
        if self.show_cursor and self.__state__:
            chars = list(self.__value__)
            chars.insert(len(self.__value__) - self.__cursor_pos, "|")
            self.update_text("".join(chars))
        else:
            self.update_text(self.__value__)
        self.__old_value__ = self.__value__

    def __manage_twe(self):
        if self.text_when_empty != None and len(self.__value__) == 0:
            self.update_text(self.text_when_empty)
            if self.auto_style:
                self.set_style(italic=True)
        else:
            if self.auto_style:
                self.set_style(self.bold_init, self.italic_init)

    def __clamp_cursor(self):
        self.__cursor_pos = max(0, min(self.__cursor_pos, len(self.__value__)))



import pygame
import os
import sys
from .Functions import vect_sum
from .Label import Label
from .Button import Button
from .Messagebox import Messagebox


class ImageFrame:
    def __init__(
            self, surface, width_height: tuple, xy: tuple, anchor="center", auto_scale=True, borderwidth=0,
            bordercolor=(0, 0, 0),
            borderradius=0):
        self.surface = surface
        self.width = width_height[0]
        self.height = width_height[1]
        self.xy = xy
        self.anchor = anchor
        self.auto_scale = auto_scale
        self.borderwidth = borderwidth
        self.bordercolor = bordercolor
        self.image = None
        self.borderradius = borderradius
        if type(self.borderradius) == int:
            self.borderradius = (
                self.borderradius, self.borderradius, self.borderradius, self.borderradius)
        elif type(self.borderradius) == tuple:
            if len(self.borderradius) != 4:
                raise ValueError(
                    f"Invalid argument for 'borderradius': {self.borderradius}.")

        self.rect = pygame.Rect(0, 0, self.width, self.height)

        # putting in correct position
        if anchor == "topleft":
            self.rect.topleft = xy
        elif anchor == "topright":
            self.rect.topright = xy
        elif anchor == "bottomleft":
            self.rect.bottomleft = xy
        elif anchor == "bottomright":
            self.rect.bottomright = xy
        elif anchor == "center":
            self.rect.center = xy
        elif anchor == "midtop":
            self.rect.midtop = xy
        elif anchor == "midright":
            self.rect.midright = xy
        elif anchor == "midbottom":
            self.rect.midbottom = xy
        elif anchor == "midleft":
            self.rect.midleft = xy

    def insert_image(self, image):
        """
        Puts a pygame image into the frame.
        """
        if self.auto_scale == True:
            self.image = pygame.transform.scale(
                image, (self.width, self.height))
        else:
            self.image = image
            self.image_rect = self.image.get_rect()
            self.image_rect.center = self.rect.center

    def draw(self):
        """
        Draws object to the screen.
        """
        if self.auto_scale == False:
            self.surface.blit(self.image, self.image_rect)
        else:
            self.surface.blit(self.image, self.rect)
        if self.borderwidth > 0:
            pygame.draw.rect(
                self.surface, self.bordercolor, self.rect, self.borderwidth,
                border_top_left_radius=self.borderradius[0],
                border_top_right_radius=self.borderradius[1],
                border_bottom_right_radius=self.borderradius[2],
                border_bottom_left_radius=self.borderradius[3])


class PlayStation_Controller_Buttons:
    def __init__(self):
        """
        Makes it easier to map PlayStation-Controller buttons to functions.
        """
        self.cross = 0
        self.circle = 1
        self.square = 2
        self.triangle = 3
        self.share_button = 4
        self.ps_button = 5
        self.options_button = 6
        self.left_stick_in = 7
        self.right_stick_in = 8
        self.l1 = 9
        self.r1 = 10
        self.arrow_up = 11
        self.arrow_down = 12
        self.arrow_left = 13
        self.arrow_right = 14
        self.touch_pad_click = 15


class Switcheroo:
    def __init__(self, surface, switcheroo_elements, previous_next_keys, lower_higher_keys,
                 scrolling_cooldown, starting_index=0, bordercolor=(255, 255, 255),
                 borderwidth=1, borderradius=0):
        """
        game = pygame game object 
        """
        self.surface = surface
        self.switcheroo_elements = switcheroo_elements
        self.previous_key, self.next_key = previous_next_keys[0], previous_next_keys[1]
        self.lower_key, self.higher_key = lower_higher_keys[0], lower_higher_keys[1]
        self.scrolling_cooldown = scrolling_cooldown
        self.scrolling_cooldown_frames = 0
        self.bordercolor = bordercolor
        self.borderwidth = borderwidth
        self.borderradius = borderradius

        self.current_index = starting_index  # the element thats currently being edited

        if starting_index not in range(len(switcheroo_elements)):
            raise ValueError(
                f"Invalid argument for 'starting index': '{starting_index}'. Index out of range.")

    def draw_box(self):
        """
        Outlines the current field.
        """
        active_element = self.switcheroo_elements[self.current_index]
        label = active_element.label_object
        pygame.draw.rect(
            self.surface, self.bordercolor,
            (label.rect.x, label.rect.y, label.rect.width, label.rect.height),
            self.borderwidth, self.borderradius)

    def update(self, keys_pressed):
        """
        Updates all values and returns them.

        Usage:
        value_1, value_2, ... = self.update(keys_pressed)
        """
        # decreases the cooldown for all elements
        for element in self.switcheroo_elements:
            if element.cooldown_frames > 0:
                element.cooldown_frames -= 1
        if self.scrolling_cooldown_frames > 0:
            self.scrolling_cooldown_frames -= 1
        # END OF BLOCK #

        # handles global scrolling first #
        keys = keys_pressed
        if not (keys[self.previous_key] and keys[self.next_key]):
            if keys[self.previous_key] and self.current_index > 0 and self.scrolling_cooldown_frames == 0:
                self.current_index -= 1
                self.scrolling_cooldown_frames = self.scrolling_cooldown
            if keys[self.next_key] and self.current_index < len(self.switcheroo_elements)-1 and self.scrolling_cooldown_frames == 0:
                self.current_index += 1
                self.scrolling_cooldown_frames = self.scrolling_cooldown
        # END OF BLOCK #

        # handles individual scrolling second #
        active_element = self.switcheroo_elements[self.current_index]
        if not (keys[self.lower_key] and keys[self.higher_key]):
            if keys[self.lower_key] and active_element.cooldown_frames == 0:
                active_element.decrease()
            if keys[self.higher_key] and active_element.cooldown_frames == 0:
                active_element.increase()
        # END OF BLOCK #

        # in the end, update all text fields and return the values
        values_list = []
        for element in self.switcheroo_elements:
            shown_value = element.var_range[element.index]
            true_value = ...

            # dict like {"Allowed":True,"Forbidden":False} ... basically {"shown_value":true_value}
            has_custom_value = False
            if element.custom_dict != None:
                if shown_value in element.custom_dict.keys():
                    true_value = element.custom_dict[shown_value]
                    has_custom_value = True

            element.label_object.update_text(shown_value)
            if has_custom_value == True:
                values_list.append(true_value)
            else:
                values_list.append(shown_value)
        # END OF BLOCK #

        return values_list


class Switcheroo_Element:
    def __init__(self, label_object, variable, var_range, starting_variable, cooldown, custom_dict=None):
        self.label_object = label_object
        self.variable = variable
        self.var_range = var_range
        self.cooldown = cooldown
        self.cooldown_frames = 0
        self.index = ...
        self.custom_dict = custom_dict

        # searching for the index of the starting variable
        # if type(starting_variable) != str:
        #     raise ValueError(f"Starting variable must be 'str'.")
        found = False
        for num, var in enumerate(var_range):
            if str(var) == str(starting_variable):
                self.index = num
                found = True
                break
        if found == False:
            raise ValueError(
                f"The starting variable '{starting_variable}' could not be found in the variables range.")

    def increase(self):
        """
        Increases self.index by 1 if its within the defined range.
        """
        if self.index + 1 < len(self.var_range):
            self.index += 1
            self.cooldown_frames = self.cooldown

    def decrease(self):
        """
        Decreases self.index by 1 if its within the defined range.
        """
        if self.index - 1 >= 0:
            self.index -= 1
            self.cooldown_frames = self.cooldown


class ImageImport_rotate:
    def __init__(self, assets_dir):
        """
        Simplifies image importing ('up','right','dowm','left').
        """
        self.assets_dir = assets_dir

    def load(self, img_name, width, height, facing="right", colorkey=None, assets_dir=None):
        if assets_dir != None:
            ad = assets_dir
        else:
            ad = self.assets_dir
        try:
            img = pygame.image.load(os.path.join(ad, img_name))
        except:
            raise ValueError(
                f"Could not find file '{img_name}' in '{ad}'.")
        img = pygame.transform.scale(img, (int(width), int(height)))
        if colorkey != None:
            img.set_colorkey(colorkey)
        if facing == "right":
            self.img_right = img
            self.img_down = pygame.transform.rotate(img, 270)
            self.img_left = pygame.transform.rotate(img, 180)
            self.img_up = pygame.transform.rotate(img, 90)
        elif facing == "down":
            self.img_down = img
            self.img_left = pygame.transform.rotate(img, 270)
            self.img_up = pygame.transform.rotate(img, 180)
            self.img_right = pygame.transform.rotate(img, 90)
        elif facing == "left":
            self.img_left = img
            self.img_up = pygame.transform.rotate(img, 270)
            self.img_right = pygame.transform.rotate(img, 180)
            self.img_down = pygame.transform.rotate(img, 90)
        elif facing == "up":
            self.img_up = img
            self.img_right = pygame.transform.rotate(img, 270)
            self.img_down = pygame.transform.rotate(img, 180)
            self.img_left = pygame.transform.rotate(img, 90)

        dct = {}
        dct["right"] = self.img_right
        dct["down"] = self.img_down
        dct["left"] = self.img_left
        dct["up"] = self.img_up
        return dct


class ImageImport_flip:
    def __init__(self, assets_dir):
        """
        Simplifies image importing ('right','left').
        """
        self.assets_dir = assets_dir

    def load(self, img_name, width, height, facing="right", colorkey=None, assets_dir=None):
        if assets_dir != None:
            ad = assets_dir
        else:
            ad = self.assets_dir
        try:
            img = pygame.image.load(os.path.join(ad, img_name))
        except:
            raise ValueError(
                f"Could not find file '{img_name}' in '{ad}'.")
        img = pygame.transform.scale(img, (int(width), int(height)))
        if colorkey != None:
            img.set_colorkey(colorkey)
        if facing == "right":
            self.img_right = img
            self.img_left = pygame.transform.flip(img, True, False)
        elif facing == "left":
            self.img_left = img
            self.img_right = pygame.transform.flip(img, True, False)

        dct = {}
        dct["right"] = self.img_right
        dct["left"] = self.img_left
        return dct


class Tile:
    def __init__(self, image, rect, type=None, hitbox=None, pos_hint=None):
        """
        image needs to be an image of the size of rect.
        rect needs to be a pygame.Rect object.
        hitbox can be None or a pygame.Rect object.
        """
        self.image = image
        self.rect = rect
        self.type = type
        self.pos_hint = pos_hint
        if hitbox != None:
            self.hitbox = pygame.Rect(
                self.rect.x + hitbox.x,
                self.rect.y + hitbox.y,
                hitbox.width,
                hitbox.height
            )


class FileDialogElement:
    def __init__(self, surface, name: str, count: int, role: str, topleft_pos, dimensions, textsize, textcolor,
                 backgroundcolor2, borderwidth):
        self.name = name
        self.role = role

        if self.role == "dir":
            bgc = (250, 150, 70)
        elif self.role == "file":
            bgc = backgroundcolor2

        self.b_name = Button(
            surface, name, textsize, (0, count * textsize),
            "topleft", fd=(dimensions[0], textsize),
            tc=textcolor, bgc=bgc, hl=True, bR=1, tb="midleft", to=(5, 1),
            aA=(topleft_pos[0] + 0, topleft_pos[1] + textsize, dimensions[0],
                dimensions[1] - textsize - 50), bc=(0, 0, 200), bw=borderwidth, br=1)

    def update(self, mouse_events, _offset):
        if self.b_name.update(mouse_events, offset=_offset):
            return True
        return False

    def draw(self):
        self.b_name.draw()


class FileDialog:
    def __init__(
            self, surface, pygame_clock, fps, topleft_pos=(100, 50),
            dimensions=(600, 400), textsize=25,
            textcolor=(0, 0, 0),
            backgroundcolor1=(150, 180, 240),
            backgroundcolor2=(128, 128, 128)):

        self.blitting_surface = surface
        self.pygame_clock = pygame_clock
        self.fps = fps
        self.topleft_pos = topleft_pos
        assert dimensions[0] >= 200
        assert dimensions[1] >= 200
        self.dimensions = dimensions
        assert 20 <= textsize <= 40
        self.textsize = textsize
        self.textcolor = textcolor
        self.backgroundcolor1 = backgroundcolor1
        self.backgroundcolor2 = backgroundcolor2


        w, h = self.dimensions
        self.center = (self.topleft_pos[0]+w//2, self.topleft_pos[1]+h//2)
        self.surface = pygame.Surface((w, h))
        self.surface.fill(self.backgroundcolor1)
        self.messagebox = Messagebox(self.blitting_surface, self.pygame_clock, self.fps, (topleft_pos[0]+dimensions[0]//2, topleft_pos[1]+dimensions[1]//2))
        self.l_path = Label(
            self.surface, "", textsize, (0, 0),
            "topleft", tc=self.textcolor, bgc=self.backgroundcolor1, br=1, tb="midright", to=(-4, 2),
            bR=1, fd=(self.dimensions[0],
                      textsize),
            bw=1)
        self.b_back = Button(
            self.surface, "<", int(textsize*1.5), (0, 0),
            "topleft", tc=self.textcolor, fd=(textsize, textsize),
            bgc=self.backgroundcolor1, hl=True, bw=1, bR=1, to=(0, -3))

        self.b_cancel = Button(self.surface, "Cancel", textsize,
                               (11, self.dimensions[1] - int(self.textsize * 1.5) // 2),
                               "midleft", tc=self.textcolor, bgc=self.backgroundcolor1, bR=1, xad=10,
                               fh=textsize, bw=3, br=1, hl=True)
        self.l_selected = Label(self.surface, "Selected:", textsize,
                                vect_sum(self.b_cancel.midright, (10, 0)),
                                "midleft", bR=1, tc=self.textcolor, bgc=self.backgroundcolor1)
        self.l_selected_name = Label(
            self.surface, "", textsize, vect_sum(self.l_selected.midright, (10, 0)),
            "midleft", bR=1, tc=self.textcolor, bgc=self.backgroundcolor1)
        self.b_open = Button(self.surface, "Open", textsize,
                             (self.dimensions[0] - 11, self.b_cancel.rect.centery),
                             "midright", tc=self.textcolor, bgc=self.backgroundcolor1, bR=1, xad=10,
                             fh=textsize, bw=3, br=1, hl=True)

        self.b_req_ending = Button(
            self.surface, "!", textsize, vect_sum(self.b_open.midleft, (-5, 0)),
            "midright", tc=(0,0,0), bR=1, f="bauhaus", to=(0, 0),
            br=1, fd=(textsize, textsize),
            bw=3, hl=True, bgc=(255, 80, 80))

    def error(self, error_message: str):
        surface = pygame.Surface((self.textsize*10, self.textsize*4))
        surface.fill(self.backgroundcolor2)
        rect = surface.get_rect()
        rect.center = self.center

        l_error = Label(surface, "Error", self.textsize, (0, 0), "topleft",
                        bR=1, tc=self.textcolor, fh=self.textsize, to=(7, 1))
        l_error_name = Label(surface, error_message, self.textsize,
                             (surface.get_width() // 2, int(self.textsize * 1.7)),
                             "midtop", tc=self.textcolor)
        b_okay = Button(
            surface, "OK", self.textsize, (surface.get_width() - 7, surface.get_height() - 7),
            "bottomright", tc=self.textcolor, bgc=self.backgroundcolor2, hl=True, xad=5, yad=3, bR=1, bw=2,
            br=1)

        run = True
        while run:
            event_list = pygame.event.get()
            mouse_events = [event for event in event_list if event.type == pygame.MOUSEBUTTONUP]
            for event in event_list:
                if event.type == pygame.QUIT:
                    pygame.quit()
                    sys.exit()
                if event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_ESCAPE:
                        run = False

            surface.fill(self.backgroundcolor2)
            l_error.draw()
            l_error_name.draw()
            b_okay.draw()
            pygame.draw.line(surface, (0, 0, 0), (0, self.textsize), (self.dimensions[0], self.textsize), 3)
            self.blitting_surface.blit(surface, rect)
            pygame.draw.rect(self.blitting_surface, (0, 0, 0), rect, 3, 1, 1, 1, 1)

            if b_okay.update(mouse_events, offset=rect.topleft):
                run = False

            pygame.display.flip()
            self.pygame_clock.tick(self.fps)

    def ask_filename(self, starting_path, required_ending: str = None, dim_background: bool = True) -> str:
        if not os.path.exists(starting_path):
            raise Exception(f"Given path does not exist. ({starting_path})")
        if required_ending != None:
            assert type(required_ending) == str, f"invalid argument for 'required_ending': {required_ending}"

        if dim_background:
            self.blitting_surface.fill([60 for i in range(3)], special_flags=pygame.BLEND_MULT)

        current_path = starting_path
        old_path = ""
        directory_scan: list

        buttons_list = []

        scroll = 0
        max_scroll = 0
        old_selected = ["", ""]  # [path, end_of_path]
        selected = ["", ""]
        self.l_selected_name.update_text("")

        run = True
        while run:
            event_list = pygame.event.get()
            mouse_events = [event for event in event_list if event.type == pygame.MOUSEBUTTONUP]
            for event in event_list:
                if event.type == pygame.QUIT:
                    pygame.quit()
                    sys.exit()
                if event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_ESCAPE:
                        run = False
                elif max_scroll > 0 and event.type == pygame.MOUSEWHEEL:
                    scroll = max(min(max_scroll, scroll + event.y*-1*int(self.textsize*1.5)), 0)

            # update
            if current_path != old_path or selected != old_selected:
                try:
                    directory_scan = os.scandir(current_path)

                    self.l_path.update_text(current_path.replace("\\", " > "))
                    old_path = current_path
                    buttons_list = []
                    MAX_ELEMENTS = 80
                    length = min(len(os.listdir(current_path)), MAX_ELEMENTS) * self.textsize
                    dir_surf = pygame.Surface((self.dimensions[0], length))
                    for count, path in enumerate(directory_scan):
                        if count == MAX_ELEMENTS:
                            break
                        new_path = os.path.join(current_path, path)
                        if os.path.isdir(new_path):
                            role = "dir"
                        elif os.path.isfile(new_path):
                            role = "file"
                        else:
                            raise Exception(f"what is this file? ({new_path})")
                        if path.name == selected[1]:
                            bw = 3
                        else:
                            bw = 0
                        buttons_list.append(FileDialogElement(dir_surf, path.name, count, role, self.topleft_pos,
                                                              self.dimensions, self.textsize, self.textcolor, self.backgroundcolor2, bw))
                    if selected == old_selected:
                        scroll = 0
                        max_scroll = length - (self.dimensions[1]-self.textsize-int(self.textsize*1.5))
                except PermissionError:
                    self.error("Permission denied.")
                    current_path = os.path.split(current_path)[0]

            if selected != old_selected:
                old_selected = selected[:]
                self.l_selected_name.update_text(selected[1])

            # updating each button
            for element in buttons_list:
                if element.update(
                    mouse_events, _offset=(self.topleft_pos[0],
                                           self.topleft_pos[1] + self.textsize + scroll)):
                    if element.role == "dir":
                        current_path = os.path.join(current_path, element.name)
                    elif element.role == "file":
                        if required_ending == None or required_ending != None and element.name.endswith(
                                required_ending):
                            selected[0] = os.path.join(current_path, element.name)
                            selected[1] = element.name

            if self.b_cancel.update(mouse_events, offset=self.topleft_pos):
                return None

            if self.b_back.update(mouse_events, offset=self.topleft_pos):
                current_path = os.path.split(current_path)[0]

            if self.b_open.update(mouse_events, offset=self.topleft_pos):
                if selected[0] != "":
                    return selected[0]
                else:
                    return None
            
            if required_ending != None and self.b_req_ending.update(mouse_events, offset=self.topleft_pos):
                self.messagebox.show_message(f"Required file ending: '{required_ending}'", False)

            # ? DRAWING ? #
            self.surface.fill(self.backgroundcolor1)

            # drawing on dir_surf
            for button in buttons_list:
                button.draw()
            self.surface.blit(dir_surf, (0, self.textsize-scroll))
            self.l_path.draw()
            self.b_back.draw()
            pygame.draw.rect(
                self.surface, (0, 0, 0),
                (0, self.textsize, self.dimensions[0],
                 self.dimensions[1] - self.textsize - int(self.textsize * 1.5)),
                1)

            # bottom part
            pygame.draw.rect(
                self.surface, self.backgroundcolor1,
                (0, self.dimensions[1] - int(self.textsize * 1.5),
                 self.dimensions[0],
                 int(self.textsize * 1.5)))
            pygame.draw.rect(
                self.surface, (0, 0, 0),
                (0, self.dimensions[1] - int(self.textsize * 1.5),
                 self.dimensions[0],
                 int(self.textsize * 1.5)),
                1)
            self.b_cancel.draw()
            self.l_selected.draw()
            self.l_selected_name.draw()
            self.b_open.draw()
            if required_ending != None:
                self.b_req_ending.draw()

            # at the end
            self.blitting_surface.blit(self.surface, self.topleft_pos)
            pygame.draw.rect(
                self.blitting_surface, (0, 0, 0),
                (self.topleft_pos[0] - 5, self.topleft_pos[1] - 5, self.dimensions[0] + 10, self.dimensions
                 [1] + 10),
                5, 1, 1, 1, 1)

            pygame.display.flip()
            self.pygame_clock.tick(self.fps)

    def ask_foldername(self, starting_path, dim_background: bool = True) -> str:
        raise Exception("not yet implemented")
